__all__ = ['phenotyping']

import numpy as np
import pickle
import os

from sklearn.metrics import hamming_loss
from sklearn.metrics import average_precision_score
from sklearn.metrics import jaccard_score
from sklearn.metrics import roc_auc_score
from sklearn.metrics import f1_score
from sklearn.metrics import coverage_error
from sklearn.metrics import label_ranking_average_precision_score
from sklearn.metrics import label_ranking_loss

def get_top_k_results(hat_y, y, k=1):
    format_hat_y = np.zeros(np.shape(hat_y))
    for i in range(len(hat_y)):
        i_s = np.argsort(hat_y[i, :])[-k:]
        for j in i_s:
            format_hat_y[i, j] = 1.

    values = {}
    values['hamming_loss'+'@'+str(k)] = hamming_loss(y, format_hat_y)
#    values['jaccard_score_micro'+'@'+str(k)] = jaccard_score(y, format_hat_y, average = 'micro')
#    values['jaccard_score_macro'+'@'+str(k)] = jaccard_score(y, format_hat_y, average = 'macro')
    values['recall'+'@'+str(k)] = np.mean(np.sum(format_hat_y * y, axis=1)/ np.sum(y, axis=1))
    values['precision'+'@'+str(k)] = np.mean(np.sum(format_hat_y * y, axis=1)/ np.sum(format_hat_y, axis=1))
    return values

def get_avg_results(hat_y, y):
    values = {}
    values['avg_precision_micro'] = average_precision_score(y, hat_y, average = 'micro')
#    values['avg_precision_macro'] = average_precision_score(y, hat_y, average = 'macro')
    values['roc_auc_score_micro'] = roc_auc_score(y, hat_y, average = 'micro')
#    values['roc_auc_score_macro'] = roc_auc_score(y, hat_y, average = 'macro')
    values['coverage_error'] = coverage_error(y, hat_y)
    values['label_ranking_average_precision_score'] = label_ranking_average_precision_score(y, hat_y)
    values['label_ranking_loss'] = label_ranking_loss(y, hat_y)
    return values

def phenotyping(hat_y, y):
    values = get_avg_results(hat_y, y)
    values_2 = get_top_k_results(hat_y, y, k=1)
    values_3 = get_top_k_results(hat_y, y, k=3)
    values.update(values_2)
    values.update(values_3)
    return values

if __name__ == '__main__':
    y = np.array([[0., 1., 0.],[1., 0., 1.]])
    hat_y = np.array([[0.3, 0.7, 0.1],[0.1, 0.2, 0.8]])
    z = phenotyping(hat_y, y)
    print (z)



import os
import csv
import pickle
import random
import numpy as np
import pandas as pd
import tqdm
from tqdm._tqdm import trange
import time
from ..utils.checklist import check_expdata_dir

class cms:

    def __init__(self, expdata_id, root_dir='.'):

        """
        experiment data generat class for cms datasets 


        Parameters

        ----------
        exp_id : str, optional (default='init.test') 
            name of current experiment
  
        """
        self.expdata_id = expdata_id
        check_expdata_dir(expdata_id =  expdata_id)
        self.root_dir = root_dir
        self.expdata_dir = os.path.join(self.root_dir, 'experiments_data', self.expdata_id)

        print(
            'Current ExpData_ID: {0} --- Target for CMS'.format(
                self.expdata_id))

    def get_exp_data(self, 
                     sel_task='phenotyping', 
                     shuffle=True, 
                     split_ratio=[0.64, 0.16, 0.2],
                     data_root = '',
                     n_limit = -1):
        """
        Parameters

        ----------
             
        task : str, optional (default='phenotyping')
            name of current healthcare task
 
        shuffle : bool, optional (default=True) 
            determine whether shuffle data or not
            
        split_ratio : list, optional (default=[0.64,0.16,0.2])
            used for split whole data into train/valid/test
        
        data_root : str, optional (default='')
            if data_root=='', use data in ./datasets; else use data in data_root
        
        n_limit : int, optional (default = -1)
            used for sample N-data not for all data, if n_limit==-1, use all data 
        """
        self.sel_task = sel_task

        if data_root == '':
            data_root = os.path.join(self.root_dir, 'datasets')

        all_list = []
        l_list = []
        episode_dir = os.path.join(data_root, 'cms', 'x_data')
        feat_n, label_n = 0, 0
        label_seq = pd.read_csv(os.path.join(data_root, 'cms', 'y_data',
                               self.sel_task + '.csv')).values
        for row_id in trange(len(label_seq)):
            if n_limit>0 and row_id>n_limit:
                break
            time.sleep(0.01)
            row = label_seq[row_id, :]
            concrete_path = os.path.join(episode_dir, row[0])
            if os.path.exists(concrete_path) is False:
                continue
            seq_l, feat_n_all = pd.read_csv(concrete_path).shape
            if seq_l < 2:
                continue
            all_list.append([concrete_path] + [seq_l] + row[1:].astype(float).tolist())
            label_n = len(row[1:])        
        feat_n = feat_n_all - 1
        # shuffle the list
        if shuffle:
            random.shuffle(all_list)
        N = len(all_list)
        x_list = []
        y_list = []
        l_list = []
        for item in all_list:
            x_list.append(item[0])
            l_list.append(item[1])
            y_list.append(np.array(item[2:]).astype(float))

        train_ratio = split_ratio[0]
        valid_ratio = split_ratio[1]

        training_x = x_list[: int(train_ratio * N)]
        validing_x = x_list[int(train_ratio * N): int(
            (train_ratio + valid_ratio) * N)]
        testing_x = x_list[int((train_ratio + valid_ratio) * N):]

        training_y = y_list[: int(train_ratio * N)]
        validing_y = y_list[int(train_ratio * N): int(
            (train_ratio + valid_ratio) * N)]
        testing_y = y_list[int((train_ratio + valid_ratio) * N):]

        training_l = l_list[: int(train_ratio * N)]
        validing_l = l_list[int(train_ratio * N): int(
            (train_ratio + valid_ratio) * N)]
        testing_l = l_list[int((train_ratio + valid_ratio) * N):]

        if os.path.exists(self.expdata_dir) is False:
            os.makedirs(self.expdata_dir)

        pickle.dump(training_x, open(
            os.path.join(self.expdata_dir, 'train_x.pkl'), 'wb'))
        pickle.dump(validing_x, open(
            os.path.join(self.expdata_dir, 'valid_x.pkl'), 'wb'))
        pickle.dump(testing_x, open(
            os.path.join(self.expdata_dir, 'test_x.pkl'), 'wb'))
        print ('finished X generate')
        pickle.dump(training_y, open(
            os.path.join(self.expdata_dir, 'train_y.pkl'), 'wb'))
        pickle.dump(validing_y, open(
            os.path.join(self.expdata_dir, 'valid_y.pkl'), 'wb'))
        pickle.dump(testing_y, open(
            os.path.join(self.expdata_dir, 'test_y.pkl'), 'wb'))
        print ('finished Y generate')
        pickle.dump(training_l, open(
            os.path.join(self.expdata_dir, 'train_l.pkl'), 'wb'))
        pickle.dump(validing_l, open(
            os.path.join(self.expdata_dir, 'valid_l.pkl'), 'wb'))
        pickle.dump(testing_l, open(
            os.path.join(self.expdata_dir, 'test_l.pkl'), 'wb'))
        print ('finished L generate')

        expdata_statistic = {
            'task':self.sel_task,
            'raio': split_ratio,
            'feat_n': feat_n,
            'label_n': label_n,
            'len_train': len(training_x),
            'len_valid': len(validing_x),
            'len_test': len(testing_x)            
        }
        pickle.dump(expdata_statistic, open(
            os.path.join(self.expdata_dir, 'expdata_statistic.pkl'), 'wb'))

        self.train = {'x': training_x, 'y': training_y, 'l': training_l,
                      'feat_n': feat_n, 'label_n': label_n}
        self.valid = {'x': validing_x, 'y': validing_y, 'l': validing_l,
                      'feat_n': feat_n, 'label_n': label_n}
        self.test = {'x': testing_x, 'y': testing_y, 'l': testing_l,
                     'feat_n': feat_n, 'label_n': label_n}

        print('generate finished')
        print('target Task:', expdata_statistic['task'])
        print('N of features:', expdata_statistic['feat_n'])
        print('N of labels:', expdata_statistic['label_n'])        
        print('N of TrainData:', expdata_statistic['len_train'])
        print('N of ValidData:', expdata_statistic['len_valid'])
        print('N of TestData:', expdata_statistic['len_test'])

    def load_exp_data(self):
        if os.path.exists(self.expdata_dir) is False:
            raise Exception('cannot find exp data dir {0}'.format(self.expdata_dir))

        training_x = pickle.load(open(
            os.path.join(self.expdata_dir, 'train_x.pkl'), 'rb'))
        validing_x = pickle.load(open(
            os.path.join(self.expdata_dir, 'valid_x.pkl'), 'rb'))
        testing_x = pickle.load(open(
            os.path.join(self.expdata_dir, 'test_x.pkl'), 'rb'))

        training_y = pickle.load(open(
            os.path.join(self.expdata_dir, 'train_y.pkl'), 'rb'))
        validing_y = pickle.load(open(
            os.path.join(self.expdata_dir, 'valid_y.pkl'), 'rb'))
        testing_y = pickle.load(open(
            os.path.join(self.expdata_dir, 'test_y.pkl'), 'rb'))

        training_l = pickle.load(open(
            os.path.join(self.expdata_dir, 'train_l.pkl'), 'rb'))
        validing_l = pickle.load(open(
            os.path.join(self.expdata_dir, 'valid_l.pkl'), 'rb'))
        testing_l = pickle.load(open(
            os.path.join(self.expdata_dir, 'test_l.pkl'), 'rb'))

        expdata_statistic = pickle.load(open(
            os.path.join(self.expdata_dir, 'expdata_statistic.pkl'), 'rb'))

        feat_n = expdata_statistic['feat_n']
        label_n = expdata_statistic['label_n']
        self.train = {'x': training_x, 'y': training_y, 'l': training_l,
                      'feat_n': feat_n, 'label_n': label_n}
        self.valid = {'x': validing_x, 'y': validing_y, 'l': validing_l,
                      'feat_n': feat_n, 'label_n': label_n}
        self.test = {'x': testing_x, 'y': testing_y, 'l': testing_l,
                     'feat_n': feat_n, 'label_n': label_n}

        print('load finished')
        print('target Task:', expdata_statistic['task'])
        print('N of features:', expdata_statistic['feat_n'])
        print('N of labels:', expdata_statistic['label_n'])        
        print('N of TrainData:', expdata_statistic['len_train'])
        print('N of ValidData:', expdata_statistic['len_valid'])
        print('N of TestData:', expdata_statistic['len_test'])

    def show_data(self, k=3):
        """
        Parameters

        ----------
        k : int, optional (default=3) 
            fetch k sample data for show
            
  
        """

        print('------------Train--------------')
        print('x_data', self.train['x'][:k])
        print('y_data', self.train['y'][:k])
        print('l_data', self.train['l'][:k])
        print('------------Valid--------------')
        print('x_data', self.valid['x'][:k])
        print('y_data', self.valid['y'][:k])
        print('l_data', self.valid['l'][:k])
        print('------------Test--------------')
        print('x_data', self.test['x'][:k])
        print('y_data', self.test['y'][:k])
        print('l_data', self.test['l'][:k])


class mimic:

    def __init__(self, expdata_id, root_dir='.'):

        """
        experiment data generat class for cms datasets 


        Parameters

        ----------
        exp_id : str, optional (default='init.test') 
            name of current experiment
  
        """
        self.expdata_id = expdata_id
        check_expdata_dir(expdata_id =  expdata_id)
        self.root_dir = root_dir
        self.expdata_dir = os.path.join(self.root_dir, 'experiments_data', self.expdata_id)

        print(
            'Current ExpData_ID: {0} --- Target for MIMIC'.format(
                self.expdata_id))

    def get_exp_data(self, 
                     sel_task='phenotyping', 
                     shuffle=True, 
                     split_ratio=[0.64, 0.16, 0.2],
                     data_root = '',
                     n_limit = -1):
        """
        Parameters

        ----------
             
        task : str, optional (default='phenotyping')
            name of current healthcare task
 
        shuffle : bool, optional (default=True) 
            determine whether shuffle data or not
            
        split_ratio : list, optional (default=[0.64,0.16,0.2])
            used for split whole data into train/valid/test
        
        data_root : str, optional (default='')
            if data_root=='', use data in ./datasets; else use data in data_root
        
        n_limit : int, optional (default = -1)
            used for sample N-data not for all data, if n_limit==-1, use all data 
        """
        self.sel_task = sel_task

        if data_root == '':
            data_root = os.path.join(self.root_dir, 'datasets')

        all_list = []
        l_list = []
        episode_dir = os.path.join(data_root, 'mimic', 'x_data')
        feat_n, label_n = 0, 0
        label_seq = pd.read_csv(os.path.join(data_root, 'mimic', 'y_data',
                               self.sel_task + '.csv')).values
        for row_id in trange(len(label_seq)):
            if n_limit>0 and row_id>n_limit:
                break
            time.sleep(0.01)
            row = label_seq[row_id, :]
            concrete_path = os.path.join(episode_dir, row[0])
            if os.path.exists(concrete_path) is False:
                continue
            seq_l, feat_n_all = pd.read_csv(concrete_path).shape
            if seq_l < 2:
                continue
            all_list.append([concrete_path] + [seq_l] + row[1:].astype(float).tolist())
            label_n = len(row[1:])        
        feat_n = feat_n_all - 1
        # shuffle the list
        if shuffle:
            random.shuffle(all_list)
        N = len(all_list)
        x_list = []
        y_list = []
        l_list = []
        for item in all_list:
            x_list.append(item[0])
            l_list.append(item[1])
            y_list.append(np.array(item[2:]).astype(float))

        train_ratio = split_ratio[0]
        valid_ratio = split_ratio[1]

        training_x = x_list[: int(train_ratio * N)]
        validing_x = x_list[int(train_ratio * N): int(
            (train_ratio + valid_ratio) * N)]
        testing_x = x_list[int((train_ratio + valid_ratio) * N):]

        training_y = y_list[: int(train_ratio * N)]
        validing_y = y_list[int(train_ratio * N): int(
            (train_ratio + valid_ratio) * N)]
        testing_y = y_list[int((train_ratio + valid_ratio) * N):]

        training_l = l_list[: int(train_ratio * N)]
        validing_l = l_list[int(train_ratio * N): int(
            (train_ratio + valid_ratio) * N)]
        testing_l = l_list[int((train_ratio + valid_ratio) * N):]

        if os.path.exists(self.expdata_dir) is False:
            os.makedirs(self.expdata_dir)

        pickle.dump(training_x, open(
            os.path.join(self.expdata_dir, 'train_x.pkl'), 'wb'))
        pickle.dump(validing_x, open(
            os.path.join(self.expdata_dir, 'valid_x.pkl'), 'wb'))
        pickle.dump(testing_x, open(
            os.path.join(self.expdata_dir, 'test_x.pkl'), 'wb'))
        print ('finished X generate')
        pickle.dump(training_y, open(
            os.path.join(self.expdata_dir, 'train_y.pkl'), 'wb'))
        pickle.dump(validing_y, open(
            os.path.join(self.expdata_dir, 'valid_y.pkl'), 'wb'))
        pickle.dump(testing_y, open(
            os.path.join(self.expdata_dir, 'test_y.pkl'), 'wb'))
        print ('finished Y generate')
        pickle.dump(training_l, open(
            os.path.join(self.expdata_dir, 'train_l.pkl'), 'wb'))
        pickle.dump(validing_l, open(
            os.path.join(self.expdata_dir, 'valid_l.pkl'), 'wb'))
        pickle.dump(testing_l, open(
            os.path.join(self.expdata_dir, 'test_l.pkl'), 'wb'))
        print ('finished L generate')

        expdata_statistic = {
            'task':self.sel_task,
            'raio': split_ratio,
            'feat_n': feat_n,
            'label_n': label_n,
            'len_train': len(training_x),
            'len_valid': len(validing_x),
            'len_test': len(testing_x)            
        }
        pickle.dump(expdata_statistic, open(
            os.path.join(self.expdata_dir, 'expdata_statistic.pkl'), 'wb'))

        self.train = {'x': training_x, 'y': training_y, 'l': training_l,
                      'feat_n': feat_n, 'label_n': label_n}
        self.valid = {'x': validing_x, 'y': validing_y, 'l': validing_l,
                      'feat_n': feat_n, 'label_n': label_n}
        self.test = {'x': testing_x, 'y': testing_y, 'l': testing_l,
                     'feat_n': feat_n, 'label_n': label_n}

        print('generate finished')
        print('target Task:', expdata_statistic['task'])
        print('N of features:', expdata_statistic['feat_n'])
        print('N of labels:', expdata_statistic['label_n'])        
        print('N of TrainData:', expdata_statistic['len_train'])
        print('N of ValidData:', expdata_statistic['len_valid'])
        print('N of TestData:', expdata_statistic['len_test'])

    def load_exp_data(self):
        if os.path.exists(self.expdata_dir) is False:
            raise Exception('cannot find exp data dir {0}'.format(self.expdata_dir))

        training_x = pickle.load(open(
            os.path.join(self.expdata_dir, 'train_x.pkl'), 'rb'))
        validing_x = pickle.load(open(
            os.path.join(self.expdata_dir, 'valid_x.pkl'), 'rb'))
        testing_x = pickle.load(open(
            os.path.join(self.expdata_dir, 'test_x.pkl'), 'rb'))

        training_y = pickle.load(open(
            os.path.join(self.expdata_dir, 'train_y.pkl'), 'rb'))
        validing_y = pickle.load(open(
            os.path.join(self.expdata_dir, 'valid_y.pkl'), 'rb'))
        testing_y = pickle.load(open(
            os.path.join(self.expdata_dir, 'test_y.pkl'), 'rb'))

        training_l = pickle.load(open(
            os.path.join(self.expdata_dir, 'train_l.pkl'), 'rb'))
        validing_l = pickle.load(open(
            os.path.join(self.expdata_dir, 'valid_l.pkl'), 'rb'))
        testing_l = pickle.load(open(
            os.path.join(self.expdata_dir, 'test_l.pkl'), 'rb'))

        expdata_statistic = pickle.load(open(
            os.path.join(self.expdata_dir, 'expdata_statistic.pkl'), 'rb'))

        feat_n = expdata_statistic['feat_n']
        label_n = expdata_statistic['label_n']
        self.train = {'x': training_x, 'y': training_y, 'l': training_l,
                      'feat_n': feat_n, 'label_n': label_n}
        self.valid = {'x': validing_x, 'y': validing_y, 'l': validing_l,
                      'feat_n': feat_n, 'label_n': label_n}
        self.test = {'x': testing_x, 'y': testing_y, 'l': testing_l,
                     'feat_n': feat_n, 'label_n': label_n}

        print('load finished')
        print('target Task:', expdata_statistic['task'])
        print('N of features:', expdata_statistic['feat_n'])
        print('N of labels:', expdata_statistic['label_n'])        
        print('N of TrainData:', expdata_statistic['len_train'])
        print('N of ValidData:', expdata_statistic['len_valid'])
        print('N of TestData:', expdata_statistic['len_test'])

    def show_data(self, k=3):
        """
        Parameters

        ----------
        k : int, optional (default=3) 
            fetch k sample data for show
            
  
        """

        print('------------Train--------------')
        print('x_data', self.train['x'][:k])
        print('y_data', self.train['y'][:k])
        print('l_data', self.train['l'][:k])
        print('------------Valid--------------')
        print('x_data', self.valid['x'][:k])
        print('y_data', self.valid['y'][:k])
        print('l_data', self.valid['l'][:k])
        print('------------Test--------------')
        print('x_data', self.test['x'][:k])
        print('y_data', self.test['y'][:k])
        print('l_data', self.test['l'][:k])




import os.path
from .base_dataset import BaseDataset
import pandas as pd
import numpy as np
import torch
import pickle

def time_series_get(fpath):
    data = pd.read_csv(fpath, sep=',')
    return data.values

class DatasetReader(BaseDataset):

    def __init__(self, data):
        BaseDataset.__init__(self, data)     
        self.series_paths = data['x']
        self.label_list = data['y']
#         self.seq_len = 200
        self.seq_len = max(data['l']) if max(data['l'])<200 else 200
        
    def __getitem__(self, index):
        xpath = self.series_paths[index]
        s_data = time_series_get(xpath)
        data = s_data[:self.seq_len, 1: ]
        l, w = np.shape(data)
        time = s_data[:self.seq_len, 0]
        time[1:] = time[1:] - time[:-1]
        time[0] = 0.
        x_time = np.zeros(self.seq_len)
        x_time[:l] = time
        x_series = np.zeros([self.seq_len, w])
        x_series[:l, :] = data
        x_mask = np.zeros(self.seq_len)
        x_mask[:l] = 1.
        x_mask_cur = np.zeros(self.seq_len)
        x_mask_cur[l-1] = 1.
        label = self.label_list[index]
        return {'X': np.array(x_series), 'M': np.array(x_mask), 'cur_M': np.array(x_mask_cur), 'Y': np.array(label), 'T':np.array(x_time)}

    def __len__(self):
        return len(self.series_paths)




# -*- coding: utf-8 -*-
# Author: Yue Zhao <zhaoy@cmu.edu>
# License: BSD 2 clause
import abc
import six


@six.add_metaclass(abc.ABCMeta)
class Standard_Template(object):
    """Abstract Class which can be inherited by various datasets,
    Key information and memory friendly information will be saved in the
    data dictionary. Otherwise, save the event and sequence location instead.
    """

    def __init__(self, patient_id):
        self.data = {}
        self.data['patient_id'] = str(patient_id)
        self.data['admission_list'] = []

    @abc.abstractmethod
    def parse_patient(self, pd_series, mapping_dict=None):
        pass

    @abc.abstractmethod
    def parse_admission(self, pd_df, mapping_dict=None):
        pass

    # @abc.abstractmethod
    def parse_icu(self, pd_df, mapping_dict=None):
        pass



# Author: Yue Zhao <zhaoy@cmu.edu>
# License: BSD 2 clause
"""MIMIC dataset handling. Adapted and modified from
https://github.com/YerevaNN/mimic3-benchmarks/blob/master/mimic3benchmark/preprocessing.py
"""
import numpy as np
import pandas as pd
import re


# Temperature: map Farenheit to Celsius, some ambiguous 50<x<80
def clean_temperature(df):
    try:
        v = df.value.astype('float32').copy()
    except ValueError:
        print("could not convert string to float in clean_temperature. "
              "Set to NaN")
        v = df.value
        v[:] = np.nan
        return v
    else:
        idx = df.valueuom.fillna('').apply(lambda s: 'F' in s.lower()) | (
                v >= 79)
        v.loc[idx] = (v[idx] - 32) * 5. / 9
        return v


def clean_sbp(df):
    v = df.value.astype(str).copy()
    idx = v.apply(lambda s: '/' in s)
    v.loc[idx] = v[idx].apply(lambda s: re.match('^(\d+)/(\d+)$', s).group(1))
    return v.astype(float)


def clean_dbp(df):
    v = df.value.astype(str).copy()
    idx = v.apply(lambda s: '/' in s)
    v.loc[idx] = v[idx].apply(lambda s: re.match('^(\d+)/(\d+)$', s).group(2))
    return v.astype(float)


def clean_crr(df):
    v = pd.Series(np.zeros(df.shape[0]), index=df.index)
    v[:] = np.nan

    # when df.value is empty, dtype can be float and comparision with string
    # raises an exception, to fix this we change dtype to str
    df_value_str = df.value.astype(str)

    v.loc[(df_value_str == 'Normal <3 secs') | (df_value_str == 'Brisk')] = 0
    v.loc[
        (df_value_str == 'Abnormal >3 secs') | (df_value_str == 'Delayed')] = 1
    return v


# Weight: some really light/heavy adults: <50 lb, >450 lb, ambiguous oz/lb
# Children are tough for height, weight
def clean_weight(df):
    try:
        v = df.value.astype('float32').copy()
    except ValueError:
        print("could not convert string to float in clean_weight. Set to NaN")
        v = df.value
        v[:] = np.nan
        return v
    else:
        # ounces
        idx = df.valueuom.fillna('').apply(lambda s: 'oz' in s.lower())
        v.loc[idx] = v[idx] / 16.
        # pounds
        idx = idx | df.valueuom.fillna('').apply(lambda s: 'lb' in s.lower())
        v.loc[idx] = v[idx] * 0.453592
        return v


# Height: some really short/tall adults: <2 ft, >7 ft)
# Children are tough for height, weight
def clean_height(df):
    try:
        v = df.value.astype('float32').copy()
    except ValueError:
        print("could not convert string to float in clean_height. Set to NaN")
        v = df.value
        v[:] = np.nan
        return v
    else:
        idx = df.valueuom.fillna('').apply(lambda s: 'in' in s.lower())
        v.loc[idx] = np.round(v[idx] * 2.54)
        return v


# FIO2: many 0s, some 0<x<0.2 or 1<x<20
def clean_fio2(df):
    try:
        v = df.value.astype('float32').copy()
    except ValueError:
        print("could not convert string to float in clean_fio2. Set to NaN")
        v = df.value
        v[:] = np.nan
        return v
    else:
        ''' The line below is the correct way of doing the cleaning, 
        since we will not compare 'str' to 'float'.
        If we use that line it will create mismatches from the data of the 
        paper in ~50 ICU stays. The next releases of the benchmark should use this line.
        '''
        # idx = df.valueuom.fillna('').apply(lambda s: 'torr' not in s.lower()) & (v>1.0)

        ''' The line below was used to create the benchmark dataset that the 
        paper used. Note this line will not work in python 3, 
        since it may try to compare 'str' to 'float'.
        '''
        # idx = df.valueuom.fillna('').apply(lambda s: 'torr' not in s.lower()) & (df.value > 1.0)

        ''' The two following lines implement the code that was used to create 
        the benchmark dataset that the paper used.
        This works with both python 2 and python 3.
        '''
        is_str = np.array(map(lambda x: type(x) == str, list(df.value)),
                          dtype=np.bool)
        idx = df.valueuom.fillna('').apply(
            lambda s: 'torr' not in s.lower()) & (
                      is_str | (~is_str & (v > 1.0)))

        v.loc[idx] = v[idx] / 100.
        return v


# GLUCOSE, PH: sometimes have ERROR as value
def clean_lab(df):
    v = df.value.copy()
    idx = v.apply(
        lambda s: type(s) is str and not re.match('^(\d+(\.\d*)?|\.\d+)$', s))
    v.loc[idx] = np.nan
    return v.astype(float)


# O2SAT: small number of 0<x<=1 that should be mapped to 0-100 scale
def clean_o2sat(df):
    # change "ERROR" to NaN
    v = df.value.copy()
    idx = v.apply(
        lambda s: type(s) is str and not re.match('^(\d+(\.\d*)?|\.\d+)$', s))
    v.loc[idx] = np.nan

    v = v.astype(float)
    idx = (v <= 1)
    v.loc[idx] = v[idx] * 100.
    return v



# -*- coding: utf-8 -*-
import os
from torch.autograd import Variable
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
import pickle
import numpy as np
import warnings
import json
import abc
import six
from ..data import rnn_reader
from ..utils.loss import loss_dict
from ..utils.checklist import *

@six.add_metaclass(abc.ABCMeta)
class BaseControler(object):
    """
    Abstract class for all healthcare predict algorithms.
    
    """
    
    @abc.abstractmethod
    def __init__(self, expmodel_id):
        """

        Parameters

        ----------
        exp_id : str, optional (default='init.test') 
            name of current experiment
            
        """
        check_model_dir(expmodel_id =  expmodel_id)
        self.checkout_dir = os.path.join('./experiments_records', expmodel_id, 'checkouts')
        self.result_dir = os.path.join('./experiments_records', expmodel_id, 'results')
        self.acc = {'train':[],'valid':[]}
        
        # make saving directory if needed
        if not os.path.isdir(self.checkout_dir):
            os.makedirs(self.checkout_dir)
            
        if not os.path.isdir(self.result_dir):
            os.makedirs(self.result_dir)

    @abc.abstractmethod
    def _build_model(self):
        pass

    def _get_device(self):
        if self.use_gpu:
            if torch.cuda.is_available():
                device = torch.device("cuda")
                print ('use GPU recource')
            else:
                device = torch.device("cpu")
                print ('not find effcient GPU, use CPU recource')
        else:
            device = torch.device("cpu")
            print ('use CPU recource')
        return device

    def _get_lossname(self, loss_name):
        if self.task in ['phenotyping']:
            if loss_name == None or loss_name == '':
                _loss_name = 'CELossSigmoid'
            elif loss_name in loss_dict['phenotyping'].keys():
                _loss_name = loss_name
            else:
                raise Exception('input correct lossfun name')
        elif self.task in ['mortality']:
            if loss_name == None or loss_name == '':
                _loss_name = 'BCELossSigmoid'
            elif loss_name in loss_dict['mortality'].keys():
                _loss_name = loss_name
            else:
                raise Exception('input correct lossfun name')
        return _loss_name

    def _get_optimizer(self):
        if self.optimizer_name == 'adam':
             return optim.Adam(self.predictor.parameters(),
                               lr=self.learn_ratio,
                               weight_decay=self.weight_decay)
    def _get_reader(self, data, dtype = 'train'):
        """
        Parameters

        ----------

        data : {
                  'x':list[episode_file_path], 
                  'y':list[label], 
                  'l':list[seq_len], 
                  'feat_n': n of feature space, 
                  'label_n': n of label space
               }

            The input samples dict.
 
        dtype: str, (default='train')
        
            dtype in ['train','valid','test'], different type imapct whether use shuffle for data
 
        Return
        
        ----------
        
        data_loader : dataloader of input data dict
        
            Combines a dataset and a sampler, and provides single- or multi-process iterators over the dataset.

            refer to torch.utils.data.dataloader
        
        """

        _dataset = rnn_reader.DatasetReader(data)
        _loader = torch.utils.data.DataLoader(_dataset,
                                              batch_size=self.n_batchsize,
                                              drop_last = True,
                                              shuffle=True if dtype == 'train' else False)
        return _loader

    def _save_predictor_config(self, predictor_config):
        temp_path = os.path.join(self.checkout_dir, "predictor_config.json")
        if os.path.exists(temp_path):
            os.remove(temp_path)
        with open(temp_path, "w", encoding='utf-8') as f:
            f.write(json.dumps(predictor_config, indent=4))

    def _load_predictor_config(self):
        temp_path = os.path.join(self.checkout_dir, 'predictor_config.json')
        assert os.path.exists(temp_path), 'cannot find predictor_config.json, please it in dir {0}'.format(self.checkout_dir)
        with open(temp_path, 'r') as f:
            predictor_config = json.load(f)
        return predictor_config

    def _save_checkpoint(self, state, epoch_id, is_best, filename='checkpoint.pth.tar'):
        if epoch_id<0:
            filepath = os.path.join(self.checkout_dir, 'latest.'+filename)
        elif is_best:
            filepath = os.path.join(self.checkout_dir, 'best.'+filename)
        else:
            filepath = os.path.join(self.checkout_dir, str(epoch_id)+'.'+filename)
        torch.save(state, filepath)

    def _train(self, train_loader):
        
        """
        Parameters

        ----------

        train_loader : dataloader of train data
        
            Combines a dataset and a sampler, and provides single- or multi-process iterators over the dataset.

            refer to torch.utils.data.dataloader

        """

        loss_v = []
        self.predictor.train()
        for batch_idx, databatch in enumerate(train_loader):
            inputs = databatch['X']
            cur_masks = databatch['cur_M']
            masks = databatch['M']
            targets = databatch['Y']
            timetick = databatch['T']
            inputs = Variable(inputs).float().to(self.device)
            masks = Variable(masks).float().to(self.device)
            cur_masks = Variable(cur_masks).float().to(self.device)
            targets = Variable(targets).float().to(self.device)
            timetick = Variable(timetick).float().to(self.device)
            data_input = {'X':inputs,'cur_M':cur_masks,'M':masks, 'T':timetick}
            all_h, h = self.predictor(data_input)
            if self.target_repl:
                data_output = {'all_hat_y': all_h, 'hat_y':h,'y':targets,'mask':masks} 
            else:
                data_output = {'hat_y':h,'y':targets}
            loss = self.criterion(data_output)
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
            loss_v.append(loss.cpu().data.numpy())
        self.acc['train'].append(np.mean(np.array(loss_v)))

    def _valid(self, valid_loader):
        """
        Parameters

        ----------

        valid_loader : dataloader of valid data
        
            Combines a dataset and a sampler, and provides single- or multi-process iterators over the dataset.

            refer to torch.utils.data.dataloader

        """

        self.predictor.eval()
        loss_v = []
        for batch_idx, databatch in enumerate(valid_loader):
            inputs = databatch['X']
            cur_masks = databatch['cur_M']
            masks = databatch['M']
            targets = databatch['Y']
            timetick = databatch['T']
            inputs = Variable(inputs).float().to(self.device)
            masks = Variable(masks).float().to(self.device)
            cur_masks = Variable(cur_masks).float().to(self.device)
            targets = Variable(targets).float().to(self.device)
            timetick = Variable(timetick).float().to(self.device)
            data_input = {'X':inputs,'cur_M':cur_masks,'M':masks, 'T':timetick}
            all_h, h = self.predictor(data_input)
            if self.target_repl:
                data_output = {'all_hat_y': all_h, 'hat_y':h,'y':targets,'mask':masks} 
            else:
                data_output = {'hat_y':h,'y':targets}
            loss = self.criterion(data_output)
            loss_v.append(loss.cpu().data.numpy())
        self.acc['valid'].append(np.mean(np.array(loss_v)))

    def _test(self, test_loader):
        """
        Parameters

        ----------

        test_loader : dataloader of test data
        
            Combines a dataset and a sampler, and provides single- or multi-process iterators over the dataset.

            refer to torch.utils.data.dataloader

        """

        # switch to train mode
        self.predictor.eval()
        pre_v = []
        prob_v = []
        real_v = []
        for batch_idx, databatch in enumerate(test_loader):
            inputs = databatch['X']
            cur_masks = databatch['cur_M']
            masks = databatch['M']
            targets = databatch['Y']
            timetick = databatch['T']
            inputs = Variable(inputs).float().to(self.device)
            masks = Variable(masks).float().to(self.device)
            cur_masks = Variable(cur_masks).float().to(self.device)
            targets = Variable(targets).float().to(self.device)
            timetick = Variable(timetick).float().to(self.device)
            data_input = {'X':inputs,'cur_M':cur_masks,'M':masks, 'T':timetick}
            _, h = self.predictor(data_input)
            if self.task in ['phenotyping']:
                prob_h = F.softmax(h, dim = -1)
            else:
                prob_h = F.sigmoid(h)
            pre_v.append(h.cpu().detach().numpy())
            prob_v.append(prob_h.cpu().detach().numpy())
            real_v.append(targets.cpu().detach().numpy())
        pickle.dump(np.concatenate(pre_v, 0), open(os.path.join(self.result_dir, 'hat_ori_y.'+self._loaded_epoch),'wb'))
        pickle.dump(np.concatenate(prob_v, 0), open(os.path.join(self.result_dir, 'hat_y.'+self._loaded_epoch),'wb'))
        pickle.dump(np.concatenate(real_v, 0), open(os.path.join(self.result_dir, 'y.'+self._loaded_epoch),'wb'))
     
    @abc.abstractmethod
    def fit(self, train_data, valid_data):
        """
        Parameters

        ----------

        train_data : {
                      'x':list[episode_file_path], 
                      'y':list[label], 
                      'l':list[seq_len], 
                      'feat_n': n of feature space, 
                      'label_n': n of label space
                      }

            The input train samples dict.
 
        valid_data : {
                      'x':list[episode_file_path], 
                      'y':list[label], 
                      'l':list[seq_len], 
                      'feat_n': n of feature space, 
                      'label_n': n of label space
                      }

            The input valid samples dict.


        Returns

        -------

        self : object

            Fitted estimator.

        """

        pass

    @abc.abstractmethod
    def load_model(self, loaded_epoch = ''):
        """
        Parameters

        ----------

        loaded_epoch : str, loaded model name 
        
            we save the model by <epoch_count>.epoch, latest.epoch, best.epoch

        Returns

        -------

        self : object

            loaded estimator.

        """

        pass

    @abc.abstractmethod
    def get_results(self):
        pass

    @abc.abstractmethod
    def inference(self, test_data):
        """
        Parameters

        ----------

        test_data : {
                      'x':list[episode_file_path], 
                      'y':list[label], 
                      'l':list[seq_len], 
                      'feat_n': n of feature space, 
                      'label_n': n of label space
                      }

            The input test samples dict.
 
 
        """

        pass



import os

def check_expdata_dir(expdata_id):
    """
    Check whether the exp data folder  exist,
        If not, will create the folder

    Parameters

    ----------
    root_dir :  str,
        root dir of current project

    expdata_id : str, optional (default='init.test') 
        name of current experiment data

    """

    r_root = os.path.join('./', 'experiments_data')
    if os.path.exists(r_root) is False:
        os.mkdir(r_root)
    exp_root = os.path.join(r_root, expdata_id)
    if os.path.exists(exp_root) is False:
        os.mkdir(exp_root)

def check_model_dir(expmodel_id):
    """
    Check whether the checkouts/results folders of current experiment(exp_id) exist,
        If not, will create both folders

    Parameters

    ----------
    root_dir :  str,
        root dir of current project

    expmodel_id : str, optional (default='init.test') 
        name of current experiment

    """

    r_root = os.path.join('./', 'experiments_records')
    if os.path.exists(r_root) is False:
        os.mkdir(r_root)
    exp_root = os.path.join(r_root, expmodel_id)
    if os.path.exists(exp_root) is False:
        os.mkdir(exp_root)
    checkout_dir = os.path.join(exp_root, 'checkouts')
    result_dir = os.path.join(exp_root, 'results')
    if os.path.exists(checkout_dir) is False:
        os.mkdir(checkout_dir)
    if os.path.exists(result_dir) is False:
        os.mkdir(result_dir)



# -*- coding: utf-8 -*-
"""A set of utility functions to support parallel computation.
"""

# Author: Yue Zhao <zhaoy@cmu.edu>
# License: BSD 2 clause
import numpy as np
from joblib import effective_n_jobs

import contextlib
import joblib
from tqdm import tqdm
from joblib import Parallel, delayed


@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    """Context manager to patch joblib to report into tqdm progress bar given as argument"""

    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()


# This elegent solution is from:
# https://stackoverflow.com/questions/24983493/tracking-progress-of-joblib-parallel-execution


def partition_estimators(n_estimators, n_jobs):
    """Private function used to partition estimators between jobs."""
    # Compute the number of jobs
    n_jobs = min(effective_n_jobs(n_jobs), n_estimators)

    # Partition estimators between jobs
    n_estimators_per_job = np.full(n_jobs, n_estimators // n_jobs,
                                   dtype=np.int)
    n_estimators_per_job[:n_estimators % n_jobs] += 1
    starts = np.cumsum(n_estimators_per_job)

    xdiff = [starts[n] - starts[n - 1] for n in range(1, len(starts))]

    # print("Split among workers default:", starts, xdiff)
    print('Split tasks into', n_jobs, 'cores...')
    return n_estimators_per_job.tolist(), [0] + starts.tolist(), n_jobs


def unfold_parallel(lists, n_jobs):
    """Internal function to unfold the results returned from the parallization

    Parameters
    ----------
    lists : list
        The results from the parallelization operations.

    n_jobs : optional (default=1)
        The number of jobs to run in parallel for both `fit` and
        `predict`. If -1, then the number of jobs is set to the
        number of cores.

    Returns
    -------
    result_list : list
        The list of unfolded result.
    """
    full_list = []
    for i in range(n_jobs):
        full_list.extend(lists[i])
    return full_list



__all__ = ['mortality']

import numpy as np
import pickle
import os

from sklearn.metrics import hamming_loss
from sklearn.metrics import average_precision_score
from sklearn.metrics import jaccard_score
from sklearn.metrics import roc_auc_score
from sklearn.metrics import f1_score
from sklearn.metrics import coverage_error
from sklearn.metrics import label_ranking_average_precision_score
from sklearn.metrics import label_ranking_loss

def get_predict_results(hat_y, y):
    _hat_y = np.array([1. if hat_y[i][0]>0.5 else 0. for i in range(len(hat_y))])
    _y = y.squeeze()
    values = {}
    values['recall'] = np.sum(_hat_y * _y)/ np.sum(_y) if np.sum(_y)>0 else 0.
    values['precision'] = np.sum(_hat_y * _y)/ np.sum(_hat_y)  if np.sum(_hat_y)>0 else 0.
    values['f1_score'] = 2 * values['precision'] * values['recall'] / (values['precision'] + values['recall']) if (values['precision'] + values['recall'])>0 else 0.
    return values

def get_avg_results(hat_y, y):
    values = {}
    values['l1'] = np.sum(abs(hat_y-y))
    values['avg_precision_micro'] = average_precision_score(y, hat_y, average = 'micro')
    values['avg_precision_macro'] = average_precision_score(y, hat_y, average = 'macro')
    values['roc_auc_score_micro'] = roc_auc_score(y, hat_y, average = 'micro')
    values['roc_auc_score_macro'] = roc_auc_score(y, hat_y, average = 'macro')
    return values

def mortality(hat_y, y):
    values = get_avg_results(hat_y, y)
    values_2 = get_predict_results(hat_y, y)
    values.update(values_2)
    return values

if __name__ == '__main__':
    y = np.array([0.,1.])
    hat_y = np.array([[0.3],[0.8]])
    cm = callMetric()
    run_type = 'train'
    z = cm.measure(hat_y, y)
    print (z)


# -*- coding: utf-8 -*-
"""A set of utility functions to support outlier detection.
"""
# Author: Yue Zhao <zhaoy@cmu.edu>
# License: BSD 2 clause

from __future__ import division
from __future__ import print_function

import os
import numpy as np
import pandas as pd
from numpy import percentile
import numbers

import sklearn
from sklearn.metrics import precision_score
from sklearn.preprocessing import StandardScaler

from sklearn.utils import column_or_1d
from sklearn.utils import check_array
from sklearn.utils import check_consistent_length

from sklearn.utils import check_random_state
from sklearn.utils.random import sample_without_replacement

MAX_INT = np.iinfo(np.int32).max
MIN_INT = -1 * MAX_INT

def make_dirs_if_not_exists(save_dir):
    # make saving directory if needed
    if not os.path.isdir(save_dir):
        os.makedirs(save_dir)

def read_csv_to_df(file_loc, header_lower=True, usecols=None, dtype=None,
                   low_memory=True, encoding=None):
    """Read in csv files with necessary processing

    Parameters
    ----------
    file_loc
    header_lower
    low_memory

    Returns
    -------

    """
    if dtype != None:
        df = pd.read_csv(file_loc, usecols=usecols, dtype=dtype,
                         low_memory=low_memory, encoding=encoding)
    else:
        df = pd.read_csv(file_loc, usecols=usecols, low_memory=low_memory,
                         encoding=encoding)

    if header_lower:
        df.columns = df.columns.str.lower()
    return df


def read_excel_to_df(file_loc, header_lower=True, usecols=None, dtype=None,
                     low_memory=True, encoding=None):
    """Read in excel files with necessary processing

    Parameters
    ----------
    file_loc
    header_lower
    low_memory

    Returns
    -------

    """
    if dtype != None:
        df = pd.read_excel(file_loc, usecols=usecols, dtype=dtype,
                           low_memory=low_memory, encoding=encoding)
    else:
        df = pd.read_excel(file_loc, usecols=usecols, low_memory=low_memory,
                           encoding=encoding)

    if header_lower:
        df.columns = df.columns.str.lower()
    return df


def check_parameter(param, low=MIN_INT, high=MAX_INT, param_name='',
                    include_left=False, include_right=False):
    """Check if an input is within the defined range.

    Parameters
    ----------
    param : int, float
        The input parameter to check.

    low : int, float
        The lower bound of the range.

    high : int, float
        The higher bound of the range.

    param_name : str, optional (default='')
        The name of the parameter.

    include_left : bool, optional (default=False)
        Whether includes the lower bound (lower bound <=).

    include_right : bool, optional (default=False)
        Whether includes the higher bound (<= higher bound).

    Returns
    -------
    within_range : bool or raise errors
        Whether the parameter is within the range of (low, high)

    """

    # param, low and high should all be numerical
    if not isinstance(param, (numbers.Integral, np.integer, np.float)):
        raise TypeError('{param_name} is set to {param} Not numerical'.format(
            param=param, param_name=param_name))

    if not isinstance(low, (numbers.Integral, np.integer, np.float)):
        raise TypeError('low is set to {low}. Not numerical'.format(low=low))

    if not isinstance(high, (numbers.Integral, np.integer, np.float)):
        raise TypeError('high is set to {high}. Not numerical'.format(
            high=high))

    # at least one of the bounds should be specified
    if low is MIN_INT and high is MAX_INT:
        raise ValueError('Neither low nor high bounds is undefined')

    # if wrong bound values are used
    if low > high:
        raise ValueError(
            'Lower bound > Higher bound')

    # value check under different bound conditions
    if (include_left and include_right) and (param < low or param > high):
        raise ValueError(
            '{param_name} is set to {param}. '
            'Not in the range of [{low}, {high}].'.format(
                param=param, low=low, high=high, param_name=param_name))

    elif (include_left and not include_right) and (
            param < low or param >= high):
        raise ValueError(
            '{param_name} is set to {param}. '
            'Not in the range of [{low}, {high}).'.format(
                param=param, low=low, high=high, param_name=param_name))

    elif (not include_left and include_right) and (
            param <= low or param > high):
        raise ValueError(
            '{param_name} is set to {param}. '
            'Not in the range of ({low}, {high}].'.format(
                param=param, low=low, high=high, param_name=param_name))

    elif (not include_left and not include_right) and (
            param <= low or param >= high):
        raise ValueError(
            '{param_name} is set to {param}. '
            'Not in the range of ({low}, {high}).'.format(
                param=param, low=low, high=high, param_name=param_name))
    else:
        return True



# -*- coding: utf-8 -*-
"""Base class for CMS dataset
"""
# Author: Yue Zhao <zhaoy@cmu.edu>
# License: BSD 2 clause
import os

import pandas as pd
import numpy as np

from .base import Standard_Template

import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter("ignore")


class CMS_Data(Standard_Template):
    """The data template to store CMS data. Customized fields can be added
    in each parse_xxx methods.

    Parameters
        ----------
        patient_id : str
            Unique identifier for a patient.
    """

    def __init__(self, patient_id, procudure_cols, diagnosis_cols):
        super(CMS_Data, self).__init__(patient_id=patient_id)
        self.procudure_cols = procudure_cols
        self.diagnosis_cols = diagnosis_cols

    def parse_patient(self, pd_series):
        self.data['gender'] = pd_series['bene_sex_ident_cd'].values[0]
        self.data['dob'] = pd_series['dob'].values[0]
        # whether there is a death date associated with it
        self.data['death_indicator'] = int(
            ~pd.isna(pd_series['bene_death_dt']).values[0])

    def parse_admission(self, pd_df, mapping_dict=None):
        # TODO: implement the mapping dict
        for ind, row in pd_df.iterrows():
            # each admission is stored as a seperate dictionary and
            # added to admission_list
            admission_event = {}
            admission_event['admission_id'] = row['clm_id']
            admission_event['admission_date'] = row['clm_from_dt']
            admission_event['discharge_date'] = row['clm_thru_dt']
            # more elements can be added here
            self.data['admission_list'].append(admission_event)

    def generate_phenotyping(self, pd_df, diagnosis_mapping_df,
                             diagnosis_codes, diagnosis_dict):
        if len(self.data['admission_list']) == 0:
            raise ValueError(
                "No admission information found. Parse admission info first.")
        # fine the last row
        last_claim = pd_df.tail(1)
        # print(last_claim)

        phenotyping_list = np.zeros([len(diagnosis_codes), 1])
        # select the procedures
        diag_df = last_claim[self.diagnosis_cols].transpose()
        diag_df.columns = ['event diagnosis code']
        # print(diag_df.shape)

        # here we join with the event mapping to receive the short code
        diag_df = diag_df.merge(diagnosis_mapping_df,
                                left_on='event diagnosis code',
                                right_on='diagnosis code cleaned')
        # print(diag_df)

        # mark the specific entry as 1 if presented
        for idx, row in diag_df.iterrows():
            # print(row['diagnosis code short'])
            phenotyping_list[diagnosis_dict[row['diagnosis code short']]] = 1

        return phenotyping_list

    def parse_event(self, pd_df, event_mapping_df, procedure_codes,
                    procedure_dict, save_dir=''):
        if len(self.data['admission_list']) == 0:
            raise ValueError(
                "No admission information found. Parse admission info first.")

        # two claims are needed for generating valid sequence
        if len(self.data['admission_list']) < 2:
            print('Only one claim find')
            return

        # make saving directory if needed
        if not os.path.isdir(save_dir):
            os.makedirs(save_dir)

        # calculate the first entry
        pd_df = pd_df.sort_values('clm_from_dt')
        first_index = pd_df['clm_from_dt'].index[0]

        pd_df['first_entry'] = pd.to_datetime(
            pd_df['clm_from_dt'][first_index])

        pd_df['days_since_entry'] = (pd_df['clm_from_dt'] - pd_df[
            'first_entry']).dt.days

        # initialize the matrix for storing the scores
        event_mat = np.zeros(
            [len(self.data['admission_list']), len(procedure_codes)])
        print(event_mat.shape)

        for i, admission_event in enumerate(self.data['admission_list']):
            print(self.data['patient_id'], admission_event['admission_id'])

            temp_df = pd_df.loc[
                pd_df['clm_id'] == admission_event['admission_id']]

            # select the procedures
            if temp_df[self.procudure_cols].shape[0] > 1:
                print('Duplicate claim ID')
                return

            proc_df = temp_df[self.procudure_cols].transpose()
            proc_df.columns = ['event procedure code']
            # print(proc_df.shape)

            # here we join with the event mapping to receive the short code
            proc_df = proc_df.merge(event_mapping_df,
                                    left_on='event procedure code',
                                    right_on='procedure code cleaned')

            # mark the specific entry as 1 if presented
            for idx, row in proc_df.iterrows():
                # print(row['procedure code short'])
                event_mat[i, procedure_dict[row['procedure code short']]] = 1

            self.data['admission_list'][i] = admission_event
        # print(event_mat, np.sum(event_mat))

        # remove all zero rows, we keep it advised by Zhi
        # event_mat = get_non_zeros_rows(event_mat)

        # skip if it is too sparse
        if np.sum(event_mat) < 2:
            print(
                'Too sparse. Fewer than two procedures in the predefined list')
            return
        # elif event_mat.shape[0] < 2:
        #     print('Only one claim left after pruning')
        #     return
        else:
            self.data['episode_csv'] = str(self.data['patient_id']) + '.csv'

            # add the timestamp
            event_mat = np.c_[pd_df['days_since_entry'], event_mat]

            pd.DataFrame(event_mat).to_csv(
                os.path.join(save_dir, self.data['episode_csv']),
                index=False, header=False)
        return



import os
import torch
import torch.nn as nn
import pickle
import warnings
from ..utils.loss import callLoss
from .base import BaseControler

warnings.filterwarnings('ignore')

class callPredictor(nn.Module):
    def __init__(self, 
                 input_size = None,
                 embed_size = 16,
                 layer_hidden_sizes = [10,20,15],
                 num_layers = 3,
                 bias = True,
                 dropout = 0.5,
                 bidirectional = True,
                 batch_first = True,
                 label_size = 1):
        super(callPredictor, self).__init__()
        assert input_size != None and isinstance(input_size, int), 'fill in correct input_size' 
        self.num_layers = num_layers
        self.rnn_models = []
        self.input_size = input_size        
        self.embed_size = embed_size
        if bidirectional:
            layer_input_sizes = [embed_size] + [2 * chs for chs in layer_hidden_sizes]
        else:
            layer_input_sizes = [embed_size] + layer_hidden_sizes
        self.label_size = label_size
        self.output_size = layer_input_sizes[-1]

        self.embed_func = nn.Linear(self.input_size, self.embed_size)
        for i in range(num_layers):
            self.rnn_models.append(nn.GRU(input_size = layer_input_sizes[i],
                                     hidden_size = layer_hidden_sizes[i],
                                     num_layers = num_layers,
                                     bias = bias,
                                     dropout = dropout,
                                     bidirectional = bidirectional,
                                     batch_first = batch_first))
        self.output_func = nn.Linear(self.output_size, self.label_size)
            
    def forward(self, input_data):
        
        """
        
        Parameters
        
        ----------
        input_data = {
                      'X': shape (batchsize, n_timestep, n_featdim)
                      'M': shape (batchsize, n_timestep)
                      'cur_M': shape (batchsize, n_timestep)
                      'T': shape (batchsize, n_timestep)
                     }
        
        Return
        
        ----------
        
        all_output, shape (batchsize, n_timestep, n_labels)
            
            predict output of each time step
            
        cur_output, shape (batchsize, n_labels)
        
            predict output of last time step

        
        """
        
        X = input_data['X']
        M = input_data['M']
        cur_M = input_data['cur_M']
        batchsize, n_timestep, n_orifeatdim = X.shape
        _ori_X = X.view(-1, n_orifeatdim)
        _embed_X = self.embed_func(_ori_X)
        _data = _embed_X.reshape(batchsize, n_timestep, self.embed_size)
        for temp_rnn_model in self.rnn_models:
            _data, _ = temp_rnn_model(_data)
        outputs = _data
        all_output = outputs * M.unsqueeze(-1)
        n_batchsize, n_timestep, n_featdim = all_output.shape
        all_output = self.output_func(outputs.reshape(n_batchsize*n_timestep, n_featdim)).reshape(n_batchsize, n_timestep, self.label_size)
        cur_output = (all_output * cur_M.unsqueeze(-1)).sum(dim=1)
        return all_output, cur_output

class EmbedGRU(BaseControler):

    def __init__(self, 
                 expmodel_id = 'test.new', 
                 task = 'phenotyping',
                 n_epoch = 100,
                 n_batchsize = 5,
                 learn_ratio = 1e-4,
                 weight_decay = 1e-4,
                 n_epoch_saved = 1,
                 embed_size = 16,
                 layer_hidden_sizes = [10,20,15],
                 bias = True,
                 dropout = 0.5,
                 bidirectional = True,
                 batch_first = True,
                 loss_name = 'L1LossSigmoid',
                 target_repl = False,
                 target_repl_coef = 0.,
                 aggregate = 'sum',
                 optimizer_name = 'adam',
                 use_gpu = False
                 ):
        """
        On an healthcare data sequence, firstly embed original features into embeded feature space, 
            then applies a multi-layer Gated recurrent unit (GRU) RNN.


        Parameters

        ----------
        exp_id : str, optional (default='init.test') 
            name of current experiment
            
        task : str, optional (default='phenotyping')
            name of current healthcare task
            
        n_epoch : int, optional (default = 100)
            number of epochs with the initial learning rate
            
        n_batchsize : int, optional (default = 5)
            batch size for model training
   
        learn_ratio : float, optional (default = 1e-4)
            initial learning rate for adam
  
        weight_decay : float, optional (default = 1e-4)
            weight decay (L2 penalty)
  
        n_epoch_saved : int, optional (default = 1)
            frequency of saving checkpoints at the end of epochs
        
        embed_size: int, optional (default = 16)
            The number of the embeded features of original input
            
        layer_hidden_sizes : list, optional (default = [10,20,15])
            The number of features of the hidden state h of each layer
            
        bias : bool, optional (default = True)
            If False, then the layer does not use bias weights b_ih and b_hh. 
            
        dropout : float, optional (default = 0.5)
            If non-zero, introduces a Dropout layer on the outputs of each GRU layer except the last layer, 
            with dropout probability equal to dropout. 

        bidirectional : bool, optional (default = True)
            If True, becomes a bidirectional GRU. 
            
        batch_first : bool, optional (default = False)
            If True, then the input and output tensors are provided as (batch, seq, feature). 
             
        loss_name : str, optional (default='SigmoidCELoss') 
            Name or objective function.

        use_gpu : bool, optional (default=False) 
            If yes, use GPU recources; else use CPU recources 

        """
 
        super(EmbedGRU, self).__init__(expmodel_id)
        self.task = task
        self.n_batchsize = n_batchsize
        self.n_epoch = n_epoch
        self.learn_ratio = learn_ratio
        self.weight_decay = weight_decay
        self.n_epoch_saved = n_epoch_saved
        self.embed_size = embed_size
        self.layer_hidden_sizes = layer_hidden_sizes
        self.num_layers = len(layer_hidden_sizes)
        self.bias = bias
        self.dropout = dropout
        self.bidirectional = bidirectional
        self.batch_first = batch_first
        self.loss_name = loss_name
        self.target_repl = target_repl
        self.target_repl_coef = target_repl_coef
        self.aggregate = aggregate
        self.optimizer_name = optimizer_name
        self.use_gpu = use_gpu
        self._args_check()
        
    def _build_model(self):
        """
        
        Build the crucial components for model training 
 
        
        """
        
        _config = {
            'input_size': self.input_size,
            'embed_size': self.embed_size,
            'layer_hidden_sizes': self.layer_hidden_sizes,
            'num_layers': self.num_layers,
            'bias': self.bias,
            'dropout': self.dropout,
            'bidirectional': self.bidirectional,
            'batch_first': self.batch_first,
            'label_size': self.label_size
            }
        self.predictor = callPredictor(**_config).to(self.device)
        self.predictor= torch.nn.DataParallel(self.predictor)
        self._save_predictor_config(_config)
        self.criterion = callLoss(task = self.task,
                                  loss_name = self.loss_name,
                                  target_repl = self.target_repl,
                                  target_repl_coef = self.target_repl_coef,
                                  aggregate = self.aggregate)
        self.optimizer = self._get_optimizer()

    def fit(self, train_data, valid_data):
        
        """
        Parameters

        ----------

        train_data : {
                      'x':list[episode_file_path], 
                      'y':list[label], 
                      'l':list[seq_len], 
                      'feat_n': n of feature space, 
                      'label_n': n of label space
                      }

            The input train samples dict.
 
        valid_data : {
                      'x':list[episode_file_path], 
                      'y':list[label], 
                      'l':list[seq_len], 
                      'feat_n': n of feature space, 
                      'label_n': n of label space
                      }

            The input valid samples dict.


        Returns

        -------

        self : object

            Fitted estimator.

        """
        self.input_size = train_data['feat_n']
        self.label_size = train_data['label_n']

        self._build_model()

        train_reader = self._get_reader(train_data, 'train')
        valid_reader = self._get_reader(valid_data, 'valid')
        
        best_score = 1e5
        for epoch in range(0, self.n_epoch):
            print('\nEpoch: [{0}|{1}]'.format(epoch + 1, self.n_epoch))
            self._train(train_reader)
            self._valid(valid_reader)
            test_score = self.acc['valid'][-1]
            print ('Train Loss : {:.3f}, Valid Loss : {:.3f}'.format(self.acc['train'][-1], self.acc['valid'][-1]))
            unit = {'epoch': epoch,
                    'state_dict': self.predictor.state_dict(),
                    'score': test_score,
                    'best_score': best_score,
                    'optimizer' : self.optimizer.state_dict()}
            if test_score<best_score:
                best_score = test_score
                unit['best_score'] = best_score
                self._save_checkpoint(unit, epoch, is_best = True)
            if epoch%self.n_epoch_saved == 0:
                self._save_checkpoint(unit, epoch, is_best = False)
            self._save_checkpoint(unit, -1, is_best = False)

    def load_model(self, loaded_epoch = ''):
        """
        Parameters

        ----------

        loaded_epoch : str, loaded model name 
        
            we save the model by <epoch_count>.epoch, latest.epoch, best.epoch

        Returns

        -------

        self : object

            loaded estimator.

        """

        predictor_config = self._load_predictor_config()
        self.predictor = callPredictor(**predictor_config).to(self.device)
        if loaded_epoch != '':
            self._loaded_epoch = loaded_epoch
        else:
            self._loaded_epoch = 'best'
        load_checkpoint_path = os.path.join(self.checkout_dir, self._loaded_epoch + '.checkpoint.pth.tar')
        if os.path.exists(load_checkpoint_path):
            try:
                checkpoint = torch.load(load_checkpoint_path)
            except:
                checkpoint = torch.load(load_checkpoint_path, map_location = 'cpu')
            self.predictor.load_state_dict({key[7:]: value for key, value in checkpoint['state_dict'].items()})
            print ('load '+self._loaded_epoch+'-th epoch model')  
        else:
            print ('no exist '+self._loaded_epoch+'-th epoch model, please dbcheck in dir {0}'.format(self.checkout_dir))

    def get_results(self):
        
        """
        
        Load saved prediction results in current ExpID
            truth_value: proj_root/experiments_records/*****(exp_id)/results/y.xxx
            predict_value: proj_root/experiments_records/*****(exp_id)/results/hat_y.xxx
            xxx represents the loaded model
        
        """
        try:
            hat_y = pickle.load(open(os.path.join(self.result_dir, 'hat_y.'+self._loaded_epoch),'rb'))
        except IOError:
            print ('Error: cannot find file {0} or load failed'.format(os.path.join(self.result_dir, 'hat_y.'+self._loaded_epoch)))
        try:
            y = pickle.load(open(os.path.join(self.result_dir, 'y.'+self._loaded_epoch),'rb'))
        except IOError:
            print ('Error: cannot find file {0} or load failed'.format(os.path.join(self.result_dir, 'y.'+self._loaded_epoch)))

        results = {'hat_y': hat_y, 'y': y}
        
        return results
  
    def inference(self, test_data):
        """
        Parameters

        ----------

        test_data : {
                      'x':list[episode_file_path], 
                      'y':list[label], 
                      'l':list[seq_len], 
                      'feat_n': n of feature space, 
                      'label_n': n of label space
                      }

            The input test samples dict.
 
 
        """
        test_reader = self._get_reader(test_data, 'test')
        self._test(test_reader)
 

    def _args_check(self):
        """
        
        Check args whether valid/not and give tips
 
        
        """
        assert isinstance(self.task,str) and self.task in ['mortality','phenotyping'], \
            'fill in correct task (str, [\'mortality\',\'phenotyping\'])'
        assert isinstance(self.n_batchsize,int) and self.n_batchsize>0, \
            'fill in correct n_batchsize (int, >0)'
        assert isinstance(self.n_epoch,int) and self.n_epoch>0, \
            'fill in correct n_epoch (int, >0)'
        assert isinstance(self.learn_ratio,float) and self.learn_ratio>0., \
            'fill in correct learn_ratio (float, >0.)'
        assert isinstance(self.weight_decay,float) and self.weight_decay>=0., \
            'fill in correct weight_decay (float, >=0.)'
        assert isinstance(self.n_epoch_saved,int) and self.n_epoch_saved>0 and self.n_epoch_saved < self.n_epoch, \
            'fill in correct n_epoch (int, >0 and <{0}).format(self.n_epoch)'
        assert isinstance(self.embed_size,int) and self.embed_size>0, \
            'fill in correct embed_size (int, >0)'
        assert isinstance(self.layer_hidden_sizes,list) and len(self.layer_hidden_sizes)>0, \
            'fill in correct layer_hidden_sizes (list, such as [10,20,15])'
        assert isinstance(self.num_layers,int) and self.num_layers>0, \
            'fill in correct num_layers (int, >0)'
        assert isinstance(self.bias,bool), \
            'fill in correct bias (bool)'
        assert isinstance(self.dropout,float) and self.dropout>0. and self.dropout<1., \
            'fill in correct learn_ratio (float, >0 and <1.)'
        assert isinstance(self.bidirectional,bool), \
            'fill in correct bidirectional (bool)'
        assert isinstance(self.batch_first,bool), \
            'fill in correct batch_first (bool)'
        assert isinstance(self.target_repl,bool), \
            'fill in correct target_repl (bool)'
        assert isinstance(self.target_repl_coef,float) and self.target_repl_coef>=0. and self.target_repl_coef<=1., \
            'fill in correct target_repl_coef (float, >=0 and <=1.)'
        assert isinstance(self.aggregate,str) and self.aggregate in ['sum','avg'], \
            'fill in correct aggregate (str, [\'sum\',\'avg\'])'
        assert isinstance(self.optimizer_name,str) and self.optimizer_name in ['adam'], \
            'fill in correct optimizer_name (str, [\'adam\'])'
        assert isinstance(self.use_gpu,bool), \
            'fill in correct use_gpu (bool)'
        assert isinstance(self.loss_name,str), \
            'fill in correct optimizer_name (str)'

        self.loss_name = self._get_lossname(self.loss_name)
        self.device = self._get_device()



import os
import torch
import torch.nn as nn
from torch.autograd import Variable

from torch.nn import L1Loss
from torch.nn import MSELoss
from torch.nn import NLLLoss
from torch.nn import BCELoss

from torch.nn import Sigmoid
from torch.nn import LogSigmoid
from torch.nn import Softmax
from torch.nn import LogSoftmax

loss_dict = {
    'phenotyping': {
        'L1LossSigmoid': {'activate': Sigmoid, 'lossfunc': L1Loss},
        'L1LossSoftmax': {'activate': Softmax, 'lossfunc': L1Loss},
        'MSELossSigmoid': {'activate': Sigmoid, 'lossfunc': MSELoss},
        'MSELossSoftmax': {'activate': Softmax, 'lossfunc': MSELoss},
        'CELossSigmoid': {'activate': LogSigmoid, 'lossfunc': BCELoss},
        'CELossSoftmax': {'activate': LogSoftmax, 'lossfunc': BCELoss}
    },
    'mortality': {
        'L1LossSigmoid': {'activate': Sigmoid, 'lossfunc': L1Loss},
        'MSELossSigmoid': {'activate': Sigmoid, 'lossfunc': MSELoss},
        'BCELossSigmoid': {'activate': Sigmoid, 'lossfunc': BCELoss}
    }
}

class callLoss(nn.Module):
        
    def __init__(self,
                 task = 'phenotyping',
                 loss_name = 'L1LossSigmoid',
                 target_repl = False,
                 target_repl_coef = 0,
                 aggregate = 'sum'):
        super(callLoss, self).__init__()
        self.loss_fn = loss_dict[task][loss_name]['lossfunc'](reduction = aggregate)
        if 'Softmax' in loss_name:
            self.activate_fn = loss_dict[task][loss_name]['activate'](dim=-1)
        else:
            self.activate_fn = loss_dict[task][loss_name]['activate']()
        self.target_repl = target_repl 
        self.target_repl_coef = target_repl_coef

    def forward(self, data):
        """
        
        Parameters
        
        ----------
        data = {
                  'hat_y': shape (batchsize, n_label)
                  
                  'y': shape (batchsize, n_label)
                  
                  'mask': [optional] shape (batchsize, n_timestep)
                      when target_repl is True
 
                  'all_hat_y': [optional] shape (batchsize, n_timestep, n_label) 
                      when target_repl is True
               }
  
  
        """
        
        hat_y, y = data['hat_y'], data['y']
        n_sample, n_label = y.size()
        y = y.view(-1, n_label)
        hat_y = hat_y.view(-1, n_label)
        hat_y = self.activate_fn(hat_y)
        single_loss = self.loss_fn(hat_y, y)
        if self.target_repl:
            all_hat_y, mask = data['all_hat_y'], data['mask']
            _, n_timestep = mask.size()
            mask_flag = mask.unsqueeze(-1)
            all_hat_y = self.activate_fn(all_hat_y) * mask_flag
            all_hat_y = all_hat_y.view(-1, n_label)
            all_y = y.unsqueeze(1).repeat(1,n_timestep,1) * mask_flag
            all_y = all_y.view(-1, n_label)
            all_loss = self.loss_fn(all_hat_y, all_y)
            loss = (1-self.target_repl_coef) * single_loss + self.target_repl_coef * all_loss
        else:
            loss = single_loss
        return loss



# -*- coding: utf-8 -*-
"""Base class for MIMIC dataset
"""
# Author: Yue Zhao <zhaoy@cmu.edu>
# License: BSD 2 clause
import os

import pandas as pd
import numpy as np

from .base import Standard_Template

from .mimic_clean_methods import clean_height, clean_weight, \
    clean_temperature, clean_dbp, clean_sbp, clean_crr, clean_lab, \
    clean_o2sat, clean_fio2

import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter("ignore")


class MIMIC_Data(Standard_Template):
    """The data template to store MIMIC data. Customized fields can be added
    in each parse_xxx methods.

    Parameters
        ----------
        patient_id

        time_duration

        selection_method
    """

    def __init__(self, patient_id, time_duration, selection_method):
        super(MIMIC_Data, self).__init__(patient_id=patient_id)
        self.time_duration = time_duration
        self.selection_method = selection_method

    def parse_patient(self, pd_series, mapping_dict=None):
        if mapping_dict is None:
            self.data['gender'] = pd_series['gender'].values[0]
            self.data['dob'] = pd_series['dob'].values[0]
        else:
            self.data['gender'] = pd_series[mapping_dict['gender']].values[0]
            self.data['dob'] = pd_series[mapping_dict['dob']].values[0]

    def parse_admission(self, pd_df):
        # TODO: implement the mapping dict
        for ind, row in pd_df.iterrows():
            # each admission is stored as a seperate dictionary and
            # added to admission_list
            admission_event = {}
            admission_event['admission_id'] = row['hadm_id']
            admission_event['admission_date'] = row['admittime']
            admission_event['discharge_date'] = row['dischtime']
            admission_event['death_indicator'] = int(
                ~pd.isna(pd_df['deathtime']).values[0])
            # more elements can be added here
            self.data['admission_list'].append(admission_event)

    def parse_icu(self, pd_df, mapping_dict=None):
        if len(self.data['admission_list']) == 0:
            raise ValueError(
                "No admission information found. Parse admission info first.")

        for i, admission_event in enumerate(self.data['admission_list']):
            # print(ind, self.data['patient_id'], admission_event['admission_id'])
            temp_df = pd_df.loc[
                pd_df['hadm_id'] == admission_event['admission_id']]
            # print(temp_df.shape)
            for ind, row in temp_df.iterrows():
                admission_event['icustay_id'] = row['icustay_id']

            self.data['admission_list'][i] = admission_event

    def parse_event(self, pd_df, save_dir='', event_mapping_df='',
                    var_list=None):

        if len(self.data['admission_list']) == 0:
            raise ValueError(
                "No admission information found. Parse admission info first.")

        # make saving directory if needed
        if not os.path.isdir(save_dir):
            os.mkdir(save_dir)

        for i, admission_event in enumerate(self.data['admission_list']):

            # get all the events
            temp_df = pd_df.loc[
                pd_df['hadm_id'] == admission_event['admission_id']]

            if not temp_df.empty:
                # save csv location
                # raw event
                admission_event['event_csv'] = str(
                    self.data['patient_id']) + '_' + str(
                    admission_event['admission_id']) + '.csv'

                # sort it by the date and time
                temp_df = temp_df.sort_values(by='charttime')

                first_index = temp_df['charttime'].index[0]
                # print(temp_df['charttime'][first_index])
                temp_df['first_entry'] = pd.to_datetime(
                    temp_df['charttime'][first_index])

                # calculate the time difference in seconds since the first entry
                temp_df['secs_since_entry'] = (temp_df['charttime'] - temp_df[
                    'first_entry']).dt.total_seconds()
                # speed up by dropping date format
                temp_df['charttime'] = temp_df['charttime'].astype(str)
                temp_df['first_entry'] = temp_df['first_entry'].astype(str)

                # save the raw event data to local csv
                temp_df.to_csv(
                    os.path.join(save_dir, admission_event['event_csv']),
                    index=False)
                # generate episode data
                # handle all events data by aggregation during a time window
                episode_df = self.generate_episode(
                    temp_df, duration=self.time_duration,
                    event_mapping_df=event_mapping_df,
                    var_list=var_list)

                if episode_df.shape[0] >= 2:
                    # save episode location only if there are more than 2 sequencnes
                    # do not save if it is empty or just a single sequence
                    admission_event['episode_csv'] = str(
                        self.data['patient_id']) + '_' + str(
                        admission_event['admission_id']) + '_episode.csv'

                    # save the episode data to local csv
                    episode_df.to_csv(
                        os.path.join(save_dir, admission_event['episode_csv']),
                        index=False)

                # update the dictionary
                self.data['admission_list'][i] = admission_event

        return temp_df  # for debug purpose

    def write_record(self, temp_list, temp_df, var):
        # TODO: this may be sped up by aggreagte on itemid, and then use the last one
        if not temp_df.empty:
            # clean specific column
            if var == 'temperature':
                temp_df['value'] = clean_temperature(temp_df)
            elif var == 'diastolic blood pressure':
                temp_df['value'] = clean_dbp(temp_df)
            elif var == 'systolic blood pressure':
                temp_df['value'] = clean_sbp(temp_df)
            elif var == 'capillary refill rate':
                temp_df['value'] = clean_crr(temp_df)
            elif var == 'weight':
                temp_df['value'] = clean_weight(temp_df)
            elif var == 'height':
                temp_df['value'] = clean_height(temp_df)
            elif var == 'fraction inspired oxygen':
                temp_df['value'] = clean_fio2(temp_df)
            elif var == 'height':
                temp_df['glucose'] = clean_lab(temp_df)
            elif var == 'ph':
                temp_df['value'] = clean_lab(temp_df)
            elif var == 'oxygen saturation':
                temp_df['value'] = clean_o2sat(temp_df)

            if self.selection_method == 'last':
                # take the last record within the range
                temp_list.append(temp_df.iloc[-1]['value'])
            elif self.selection_method == 'mean':
                temp_list.append(temp_df['value'].mean())
                # if there is no information, just skip
        else:
            temp_list.append('')

        return temp_list

    def generate_episode_headers(self, var_list):
        """Generate the header for episode file

        Parameters
        ----------
        var_list

        Returns
        -------

        """

        self.episode_headers_ = ['secs_since_entry']
        for var in var_list:
            # self.episode_headers_.extend([var, var + '_unit'])
            self.episode_headers_.append(var)

    def generate_episode(self, pd_df, duration, event_mapping_df, var_list):

        self.generate_episode_headers(var_list)

        # when it is in the memory, process it directly
        max_time_diff = pd_df['secs_since_entry'].max()

        # only keep the events we are interested in
        key_df = event_mapping_df[event_mapping_df['level2'].isin(var_list)]

        # rounding up
        n_episode = int(np.ceil(max_time_diff / duration))

        # print(n_episode)   

        episode_df = pd.DataFrame(columns=self.episode_headers_)

        for j in range(n_episode):
            threshold_l = j * duration
            threshold_h = (j + 1) * duration

            # find all the events within the duration
            slice_df = pd_df.loc[(pd_df['secs_since_entry'] >= threshold_l) & (
                    pd_df['secs_since_entry'] < threshold_h)]

            # need some sort on the time so just get the last value within the range
            # can join on multiple things, weight for instance
            temp_df = key_df.merge(slice_df, left_on='itemid',
                                   right_on='itemid')

            # valid information is available, otherwise skip            
            if not temp_df.empty:
                temp_df.sort_values(by='secs_since_entry', inplace=True)

                # initialize the record with timestamp
                temp_list = [threshold_h]

                # iterate over variables at interests
                for var in var_list:
                    var_df = temp_df[temp_df['level2'] == var]
                    # write records for each variable at interest
                    temp_list = self.write_record(temp_list, var_df, var)

                # append to the major episode dataframe
                epi_df = pd.DataFrame(temp_list).transpose()
                epi_df.columns = self.episode_headers_
                episode_df = pd.concat([episode_df, epi_df], axis=0)

                # ###############################################################
                # # need to iterate over rows, may not be sufficiently faster.
                # # on park
                # get the last record for each variable
                # episode_df = temp_df.groupby(['level2']).agg({
                #     'itemid' : 'last',
                #     'value' : 'last',
                #     'valueuom' : 'last',
                #     })                
                # self.a = episode_df
                # print(episode_df)
                # ###############################################################

        # return the episode df 
        return episode_df


##############################################################################

def parallel_parse_tables(patient_id_list, patient_df, admission_df, icu_df,
                          event_df, event_mapping_df, duration,
                          selection_method, var_list, save_dir):
    """Parallel methods to process patient information in batches

    Parameters
    ----------
    patient_id_list
    patient_df
    admission_df
    icu_df
    event_df
    var_list

    Returns
    -------

    """
    valid_data_list, valid_id_list = [], []

    # for i in tqdm(range(len(patient_id_list))):
    for i in range(len(patient_id_list)):
        p_id = patient_id_list[i]
        # print('Processing Patient', i + 1, p_id)
        # initialize the 
        temp_data = MIMIC_Data(p_id, duration, selection_method)
        p_df = patient_df.loc[patient_df['subject_id'] == p_id]
        a_df = admission_df.loc[admission_df['subject_id'] == p_id]
        i_df = icu_df.loc[icu_df['subject_id'] == p_id]
        e_df = event_df.loc[event_df['subject_id'] == p_id]

        if not p_df.empty:
            if p_df.shape[0] > 1:
                raise ValueError("Patient ID cannot be repeated")
            temp_data.parse_patient(p_df)

        if not a_df.empty:
            temp_data.parse_admission(a_df)
        if not i_df.empty:
            temp_data.parse_icu(i_df)

        if not e_df.empty:
            temp_data.parse_event(e_df, save_dir=save_dir,
                                  event_mapping_df=event_mapping_df,
                                  var_list=var_list)

        valid_data_list.append(temp_data)
        valid_id_list.append(p_id)

    return valid_data_list, valid_id_list



import random
import numpy as np
import torch.utils.data as data
from abc import ABC, abstractmethod
import torch


class BaseDataset(data.Dataset, ABC):

    def __init__(self, opt):
        self.opt = opt

    @staticmethod
    def modify_commandline_options(parser, is_train):
        return parser

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def __getitem__(self, index):
        pass
